from tkinter import *
import math
import copy
import types
def array(N1,N2):
    return [[0 for i in range(N2)] for j in range(N1)]
class Point:
    def __init__(self, X, Y, Z):    #コンストラクタ
        self.x=copy.copy(X)
        self.y=copy.copy(Y)
        self.z=copy.copy(Z)
    def __add__(self, P):           # +
        return Point(self.x+P.x, self.y+P.y, self.z+P.z)
    def __sub__(self, P):           # -(減算)
        return Point(self.x-P.x, self.y-P.y, self.z-P.z)
    def __neg__(self):              # -(マイナス)
        return Point(-self.x, -self.y, -self.z)
    def __mul__(self, P):           # *
        if isinstance(P,float) or  isinstance(P,int) :
            return Point(self.x*P, self.y*P, self.z*P)
        elif isinstance(P,Point):
            return self.x*P.x+self.y*P.y+self.z*P.z
        else:
            print("この型はサポートしていません。",type(P))
            return 0
    def __mod__(self, P):           # %は使わないので外積用演算子として用いる
        if isinstance(P,float) or  isinstance(P,int) :
            return Point(self.x*P, self.y*P, self.z*P)
        elif isinstance(P,Point):
            return Point(self.y*P.z-self.z*P.y,
                         self.z*P.x-self.x*P.z,
                         self.x*P.y-self.y*P.x)
        else:
            print("この型はサポートしていません。",type(P))
            return Point(self.x,self.y, self.z)
    def __truediv__(self, P):       #  /
        if isinstance(P,float) or  isinstance(P,int) :
            return Point(self.x/P, self.y/P, self.z/P)
        else:
            print("この型はサポートしていません。",type(P))
            return Point(self.x,self.y, self.z)
    def length(self):               #ベクトルの長さ
        return math.sqrt(self.x*self.x+self.y*self.y+self.z*self.z)
    def __eq__(self,P):             #  ==
        return self.x==P.x and self.y==P.y and self.z==P.z
    def __ne__(self,P):             #  !=
        return not(self==P)
    def print(self,form="x= %8.4f, y= %8.4f, z=%8.4f"):
        print(form % (self.x, self.y, self.z))
#■単位ベクトル(Pがベクトルのとき、P方向の単位ベクトル，eは無視。
#               Pが数値のとき、P:方位角, e：仰角の単位ベクトル)
def unit(P, e=0):
    if isinstance(P,Point):
        if P.length()==0: return Point(1,0,0)
        return P / P.length()
    if isinstance(P,float) or  isinstance(P,int) :
        PIR=math.pi/180; aa = P * PIR; ee = e * PIR
        X=math.cos(aa)*math.cos(ee)
        Y=math.sin(aa)*math.cos(ee)
        Z=math.sin(ee)
        return Point(X,Y,Z)
    else:
        print("この型はサポートしていません。",type(P))
        return P
#■Lineクラス
class Line:
    def __init__(self,Plist):
        if isinstance(Plist,Line):
            self.p=copy.copy(Plist.p)
            return
        self.p=[]; N=Plist; iflag=True
        if isinstance(Plist,list) or isinstance(Plist,tuple) :
            N=len(Plist); iflag =False
       
        for i in range(N):
            if iflag:  self.p.append(0)
            else: self.p.append(copy.copy(Plist[i]))

    def print(self,form="x= %8.4f, y= %8.4f, z=%8.4f"):
        for i in range(len(self.p)):
            print((" %4d : "+ form) % (i, self.p[i].x, self.p[i].y, self.p[i].z))
            
    def __mul__(self, P):
        C=Line(self.p)
        if isinstance(P,float) or  isinstance(P,int) :
            for i in range(len(self.p)):
                C.p[i].x *= P; C.p[i].y *= P; C.p[i].z *= P 
        else:
            print("この型はサポートしていません。",type(P))
        return C
    def length(line):
        LN=0
        for i in range(len(line.p)-1):
            LN+= (line.p[i] - line.p[i+1]).length()
        return LN
#■ベクトル/線の長さ(分かりやすくするためこれを用意)
def length(P):
    if isinstance(P,Point) or isinstance(P,Line): return P.length()
    else:
        print("この型はサポートしていません。",type(P))
        return 0
#■Polygonクラス
class Polygon(Line):
    def __init__(self,Plist):
        if isinstance(Plist,Polygon):
            self.p=copy.copy(Plist.p)
            return
        self.p=[]; N=Plist; iflag=True
        if isinstance(Plist,Line):
            Plist=Plist.p
        if isinstance(Plist,list) or isinstance(Plist,tuple) :
            N=len(Plist); iflag =False
        for i in range(N):
            if iflag:  self.p.append(0)
            else: self.p.append(copy.copy(Plist[i]))
        self.p.append(self.p[0])
    def print(self,form="x= %8.4f, y= %8.4f, z=%8.4f"):
        for i in range(len(self.p)):
            print((" %4d : "+ form) % (i, self.p[i].x, self.p[i].y, self.p[i].z))
#■TMatrixクラス            
class TMatrix(Point):
    def __init__(self,Plist=0):
        self.t=array(4,4)
        for i in range(4):
            self.t[i][i]=1
        if Plist==0: return
        N1=len(Plist)
        if N1>4: N1=4
        for i in range(N1):
            N2=len(Plist[i])
            if N2>4: N2=4
            for j in range(N2):
                self.t[i][j]=copy.copy(Plist[i][j])
    def print(self,form="%8.4f, %8.4f, %8.4f, %8.4f"):
        for i in range(len(self.t)):
            print((" %3d : "+ form) %
                  (i, self.t[i][0], self.t[i][1], self.t[i][2], self.t[i][3]))
    def __mul__(self, b):           # *
        if isinstance(b,float) or  isinstance(b,int) :
            #t=TMatrix(self.t)
            #for i in range(4):
            #    for j in range(4):
            #        t.t[i][j]*=b
            print(type(self), type(scale(b)))
            return self*scale(b)
        elif isinstance(b,Line):
            c=Line(b)
            for i in range(len(c.p)):
                c.p[i]= self*b.p[i]
            return c
        elif isinstance(b,Polygon):
            c=Polygon(b)
            for i in range(len(c.p)):
                c.p[i]= self*b.p[i]
            return c
        elif isinstance(b,Surface):
            c=Surface(len(b.polygon))
            for i in range(len(c.polygon)):
                c.polygon[i]=self*b.polygon[i]
            return c

        elif isinstance(b,TMatrix):
            c=TMatrix()
            for i in range(4):
                for j in range(4):
                    c.t[i][j]=0
                    for k in range(4):
                        c.t[i][j] += self.t[i][k]*b.t[k][j]
            return c
        elif isinstance(b, Point):
            return Point(self.t[0][0]*b.x+self.t[0][1]*b.y+self.t[0][2]*b.z+self.t[0][3],
                         self.t[1][0]*b.x+self.t[1][1]*b.y+self.t[1][2]*b.z+self.t[1][3],
                         self.t[2][0]*b.x+self.t[2][1]*b.y+self.t[2][2]*b.z+self.t[2][3])
        else:
            print("この型はサポートしていません。",type(P))
            return 0
def scale(b):           # scale
    S=TMatrix()
    if isinstance(b,float) or  isinstance(b,int) :
        sx=b; sy=b; sz=b
    elif isinstance(b,Point):
        sx=b.x; sy=b.y; sz=b.z
    elif isinstance(b,list) or isinstance(b,tuple) :
        sx=b[0]; sy=b[1]; sz=b[2]
    else:
        print("この型はサポートしていません。",type(P))
        sx=1; sy=1; sz=1
    S.t[0][0]=sx; S.t[1][1]=sy; S.t[2][2]=sz; S.t[3][3]=1
    return S
#■回転
def rot_x(r): #　Ｘ軸回り
    m=TMatrix()
    if isinstance(r,float) or  isinstance(r,int) :
        PIR=math.pi*r/180
        sn=math.sin(PIR); cs=math.cos(PIR)
    elif isinstance(r,list) or isinstance(r,tuple) :
        sn=r[0]; cs=r[1]
    else:
        print("この型はサポートしていません。",type(P))
        sn=0; cn=1
    m.t[1][1]=cs; m.t[1][2]=-sn;m.t[2][1]=sn; m.t[2][2]=cs
    return m
def rot_y(r): #　Ｙ軸回り
    m=TMatrix()
    if isinstance(r,float) or  isinstance(r,int) :
        PIR=math.pi*r/180
        sn=math.sin(PIR); cs=math.cos(PIR)
    elif isinstance(r,list) or isinstance(r,tuple) :
        sn=r[0]; cs=r[1]
    else:
        print("この型はサポートしていません。",type(P))
        sn=0; cn=1
    m.t[0][0]=cs; m.t[0][2]=sn;m.t[2][0]=-sn; m.t[2][2]=cs
    return m
def rot_z(r): #　Z軸回り
    m=TMatrix()
    if isinstance(r,float) or  isinstance(r,int) :
        PIR=math.pi*r/180
        sn=math.sin(PIR); cs=math.cos(PIR)
    elif isinstance(r,list) or isinstance(r,tuple) :
        sn=r[0]; cs=r[1]
    else:
        print("この型はサポートしていません。",type(P))
        sn=0; cn=1
    m.t[0][0]=cs; m.t[0][1]=-sn;m.t[1][0]=sn; m.t[1][1]=cs
    return m
def rot_xz(a): # X軸回りに回転し, Z軸回りに回転(a:Point）
    ua=unit(a)
    aa=scale((1,1,0))*ua
    lenAA=length(aa)
    if lenAA==0 : return rot_x(90)
    mx=rot_x((Point(0,0,1)*ua, lenAA))
    mz=rot_z((-aa.x/lenAA, aa.y/lenAA))
    return mz*mx
def rotate(a): #　Ｙ軸回り
    m=TMatrix()
    if isinstance(a,float) or  isinstance(a,int) : return rot_z(a)
    if isinstance(a,Point) : #ベクトル方向に回転
        P  = Point(1,0,0)
        ss = P % unit(a)
        if length(ss)==0 : ss= Point(0,0,1)
        s  = unit(ss); t1 = rot_x((-s.y,s.z))
        t3 = rot_x((s.y,s.z))
        m  = unit(t1*a)
        t2 = rot_z((m.y, m.x))
        return  t3*t2*t1;
    iflg=False
    if isinstance(a,list) or isinstance(a,tuple) :
        if len(a)==2:
            if isinstance(a[0],Point) and isinstance(a[1],Point):
                sn=length(unit(a[0]) % unit(a[1]))
                if sn ==0: return TMatrix()
                cs=unit(a[0])*unit(a[1])
                aa=unit(a[0]) % unit(a[1])
                iflg=True
            elif isinstance(a[0],Point) and (isinstance(a[1],float) or  isinstance(a[1],int)):
                aa=a[0]
                PIR=math.pi*a[1]/180
                sn=math.sin(PIR); cs=math.cos(PIR)
                iflg=True
        elif len(a)==3:
            if (isinstance(a[0],Point) and
                (isinstance(a[1],float) or  isinstance(a[1],int)) and
                (isinstance(a[2],float) or  isinstance(a[2],int))):
                   aa=a[0]; sn=a[1]; cs=a[2]
                   iflg=True
    if iflg:
        P=Point(aa.x, aa.y,0)
        axy=unit(P)
        axz=rot_z((-axy.y, axy.x))*unit(aa)
        return (rot_z((axy.y, axy.x))*rot_y((-axz.z, axz.x))*rot_x((sn,cs))*
                rot_y((axz.z, axz.x))*rot_z((-axy.y, axy.x)))
    else:
        print("この型はサポートしていません。",type(a))
        return TMatrix()
#■移動行列
def move(P, y=0, z=0):
    if isinstance(P,Point):
        x=P.x; y=P.y; z=P.z
    else: x=P
    m=TMatrix()
    m.t[0][3]=x; m.t[1][3]=y; m.t[2][3]=z
    return m
#■透視変換
def perspect(d):
    S=TMatrix()
    S.t[2][3] = -d; S.t[3][2] = -1/d
    return S
#■転置行列
def trans(a):
    c=TMatrix()
    for i in range(4):
        for j in range(4):
            c.t[i][j]=a.t[j][i]
    return c
#■逆変換行列
def inverse(a):
    d=0;
    for i in range(3):
        d = (d + a.t[0][i]*a.t[1][(i+1)%3]*a.t[2][(i+2) % 3]
             - a.t[0][i]*a.t[1][(i+2)%3]*a.t[2][(i+1) % 3])
    b=TMatrix()
    if d != 0:
        for i in range(3):
            for j in range(3):
                b.t[i,j]=(a.t[(j+1)%3][(i+1)%3]*a.t[(j+2)%3][(i+2)%3]
                          -a.t[(j+1)%3][(i+2)%3]*a.t[(j+2)%3][(i+1)%3])/d
    for i in range(3): b.t[i][3] = -a.t[i][3]
    return b;
#■Surfaceクラス
class Surface:
    def __init__(self,a=0):
        if isinstance(a,int) :
            self.polygon=[0 for i in range(a)]
        elif isinstance(a,Polygon):
            self.polygon=[copy.copy(a)]
        elif isinstance(a,Surface):
            self.polygon=[]
            for i in range(len(a)):
                self.polygon.append(copy.copy(a[i]))
        else:
            print("この型はサポートしていません。",type(P))
    def print(self,form="%8.4f, %8.4f, %8.4f"):
        for i in range(len(self.polygon)):
            print(" %d 番目Polygon" % i)
            P=self.polygon[i].p
            for j in range(len(P)):
                print((  " %3d : "+ form) %
                      (j, P[j].x, P[j].y, P[j].z))
#■線分aをｚ軸回転してポリゴンモデルを作成                
def revolve_z(a, n, b=0, e=360):
    f=Surface((len(a.p)-1)*n)
    t=0
    for j in range(n):
        m0=rot_z(b+(e-b)*j/n)
        m1=rot_z(b+(e-b)*(j+1)/n)
        for i in range(len(a.p)-1):
            f.polygon[t]=Polygon((m0*a.p[i],m0*a.p[i+1],m1*a.p[i+1],m1*a.p[i]))
            t+=1
    return f
#■線分aを線分bに沿ってポリゴンモデルを作成
def sweep_xz(a, b):
    n=len(b.p)
    na=len(a.p)
    f=[0 for i in range(n)]
    if(length(b.p[0]-b.p[n-1])>0.1):
        f[0]=move(b.p[0])*rot_xz(b.p[1]-b.p[0])*a
        f[n-1]=move(b.p[n-1])*rot_xz(b.p[n-1]-b.p[n-2])*a
    else:
        f[0]    =move(b.p[0])*rot_xz(b.p[1]-b.p[n-4])*a
        f[n-1]=move(b.p[0])*rot_xz(b.p[3]-b.p[n-2])*a
    for i in range(n-1):
        f[i]=move(b.p[i])*rot_xz(b.p[i+1]-b.p[i-1])*a
    s=Surface((na-1)*(n-1))
    t=0
    for i in range(n-1):
        for j in range(na-1):
            s.polygon[t]=Polygon((f[i].p[j],f[i].p[j+1],f[i+1].p[j+1],f[i+1].p[j]));
            t+=1
    return s
#■原点の設定
org=Point(0,0,0)
def origin(x, y):
    global org
    org.x=x+0.5; org.y=y-0.5; org.z=0
#■視点方向設定
v0=Point(0,0,0)
dv=100000
t=TMatrix()
it=TMatrix()
def setview(a, e=0, d=100000):
    global v0, t,it,dv
    P1=Point(0,0,1);
    if isinstance(a,Point) : #ベクトル方向に回転
        dv=length(a)
        wa=unit(a)
        if (wa.x==0) and (wa.y==0) :
            t=perspect(dv)
            it=copy(t)
        else:
            wd=math.sqrt(wa.x*wa.x+wa.y*wa.y)
            
            t=perspect(dv)*rot_x((-wd,wa.z))*rot_z((-wa.x/wd,-wa.y/wd))
            it=rot_z((wa.x/wd,-wa.y/wd))*rot_x((wd,wa.z))
    else:
        dv=d
        t=perspect(d)*rot_x(e-90)*rot_z(-90-a)
        it=rot_z(90+a)*rot_x(90-e)
        v0=it*P1*d
def world(p):
    global v0, t,it,dv
    sz=-p.z/dv
    return it*Point((p.x-org.x)*sz,(-p.y+org.y)*sz, p.z+dv)
#■スクリーン座標への変換
def screen(p):
    a=t*p
    s=t.t[3][0]*p.x+t.t[3][1]*p.y+t.t[3][2]*p.z+t.t[3][3]
    return Point(org.x+a.x/s,org.y-a.y/s,a.z)
#■初期設定
size_xy=Point(0,0,0)
def init2D(x,y):
    global size_xy
    size_xy.x=x
    size_xy.y=y
    origin(size_xy.x/2-0.5, size_xy.y/2-1.5)
    setview(30,30,1000)
#■線描画(３Ｄ座標系)
def drawLine(canvas, LDT, color):
    if isinstance(LDT,Surface):
        for i in range(0,len(LDT.polygon)):
            drawLine(canvas, LDT.polygon[i],color)
        return
    if isinstance(LDT,Line) or isinstance(LDT,Polygon):
        LL=LDT.p
    elif isinstance(LDT,list) or isinstance(LDT,tuple):
        LL=LDT
        for i in range(1,len(LDT.p)):
            P2=LDT[i]
            PP1 = screen(P1) #標準座標からスクリーン座標への変換
            PP2 = screen(P2)
            canvas.create_line(PP1.x, PP1.y, PP2.x, PP2.y, fill=color, tag="3D");
            P1 = P2
    else:
        print("この型はサポートしていません。",type(LDT))
        return
    P1=LL[0]
    for i in range(1,len(LDT.p)):
        P2=LL[i]
        PP1 = screen(P1) #標準座標からスクリーン座標への変換
        PP2 = screen(P2)
        canvas.create_line(PP1.x, PP1.y, PP2.x, PP2.y, fill=color, tag="3D");
        P1 = P2
#■点描画(スクリーン座標系)
def drawPointXY(canvas, x,y,C):
    canvas.create_rectangle(x, y, x, y,fill=C,outline='', tag="3D")
#■点描画(３Ｄ座標系)
def drawPoint(canvas, p, C):
    b=screen(p)
    drawPointXY(canvas, b.X,b.Y,C)
#■多角形描画(３Ｄ座標系)
def paintPolygon(canvas, p,C):
    a=[]
    for i in range(len(p.p)-1):
        b=screen(p.p[i])
        a.append(b.x)
        a.append(b.y)
    canvas.create_polygon(a, fill=C, outline=C, tag="3D")
#■軸の描画
def drawAxis(canvas, AX, font, P1, P2, color):
    PP1 = screen(P1) #標準座標からスクリーン座標への変換
    PP2 = screen(P2)
    canvas.create_line(PP1.x, PP1.y, PP2.x, PP2.y, fill=color, tag="3D");
    canvas.create_text(PP2.x+10,PP2.y, text=AX,fill=color, font=font, tag="3D")
#■XYZ軸描画
def drawAxisXYZ(canvas):
    font =('Times',10)
    drawAxis(canvas, "X",font, Point(-size_xy.x/2, 0, 0), Point(size_xy.x/2, 0, 0),'black')
    drawAxis(canvas,"Y",font, Point(0, -size_xy.y/2, 0), Point(0, size_xy.y/2, 0),'black')
    drawAxis(canvas,"Z",font, Point(0, 0, -size_xy.y/2), Point(0, 0, size_xy.y/2),'black')
#■折れ線
def circle_xz(n, b=0, e=360):# y = 0 (xz平面上の折れ線)
    c=Line(n+1)
    PIR=math.pi/180
    TH=b*PIR
    DTH=PIR*(e-b)/n
    for i in range(n+1):
        c.p[i]=Point(math.sin(TH),0, math.cos(TH))
        TH += DTH
    return c
def circle_yz(n, b=0, e=360):# x = 0 (yz平面上の折れ線)
    c=Line(n+1)
    PIR=math.pi/180
    TH=b*PIR
    DTH=PIR*(e-b)/n
    for i in range(n+1):
        c.p[i]=Point(0,math.cos(TH),math.sin(TH))
        TH += DTH
    return c;
def circle_xy(n, b=0, e=360):# z = 0 (xy平面上の折れ線)
    c=Line(n+1)
    PIR=math.pi/180
    TH=b*PIR
    DTH=PIR*(e-b)/n
    for i in range(n+1):
        c.p[i]=Point(math.cos(TH),math.sin(TH),0)
        TH += DTH
    return c;
#■補間(p1,p2を補間)
def interpolate(p1, p2, a1,a2):#p1,p2を補間
    a12=a1+a2
    if a12<1: return p1
    else :return (p1*a2+p2*a1)/(a1+a2)
#■カラークラス
class Color:
    def __init__(self, R, G, B):
        self.R=R
        self.G=G
        self.B=B
    def __add__(self,C):
        return Color(self.R+C.R, self.G+C.G, self.B+C.B)
    def __mul__(self,C):
        return Color(int(self.R*C), int(self.G*C), int(self.B*C))
    def __truediv__(self,C):
        return Color(int(self.R/C), int(self.G/C), int(self.B/C))
        
    def toString(self):
        return '#%02X%02X%02X' % (int(self.R), int(self.G),int(self.B)) 
def mixValue(D1, V1, D2, V2):
    R=int(D1*float(V1)+D2*float(V2))
    if R>255: return 255
    if R<0  : return 0
    return R
def shading(V, L, N, CU, s=1, kd=0.7, ks=0.7, ke=0.3):
    # kd=拡散反射係数, ks=鏡面反射係数, ke=環境光係数;
    NL=-N*L*s
    if NL<0: NL=0
    RR=L-N*(L*N*2);  RV=-RR*V                  #反射方向ベクトル
    if RV<0: RV=0
    D2=ks*math.pow(RV,20)*s
    D1=NL*kd+ke
    R=int(mixValue(D1, CU.R, D2, 255))
    G=int(mixValue(D1, CU.G, D2, 255))
    B=int(mixValue(D1, CU.B, D2, 255))
    return Color(R,G,B)
def rangeScreen(pp,Num):
    y_max=int(pp[0].y); y_min=int(pp[0].y)    #スクリーン座標範囲
    s0=0; s1=0; 
    for i in range(Num):
        if y_min > int(pp[i].y): y_min=pp[i].y; s0=i; s1=i
        elif y_min==int(pp[i].y): s1=i
        if int(pp[i].y)>y_max: y_max=int(pp[i].y)
    return [y_min, y_max, s0, s1]
def setSID(pp, s, n, y):
    S=[0,0,0,0]; S[0]=s[0]; S[1]=s[1]
    if pp[(s[0]-1+n) % n].y < y : S[0]=(s[0]-1+n) % n
    if pp[(s[1]+1)%n].y < y     : S[1]=(s[1]+1) %n
    S[2]=(S[1]+1) %n; S[3]=(S[0]-1+n) % n
    return S
def interPos(p1, p2, y): #位置の補間
    return interpolate(p1, p2, y-int(p1.y), int(p2.y)-y)
def checkRange(D, D1,D2): #範囲チェック
    if D<D1 : return D1
    if D>D2 : return D2
    return D
def interCol(c1, c2, a1, a2):#カラーの補間
    C=interpolate(c1,c2,int(a1),int(a2))
    C.R=checkRange(C.R, 0, 255)
    C.G=checkRange(C.G, 0, 255)
    C.B=checkRange(C.B, 0, 255)
    return C
def mix(C1,d1,C2,d2):
    R= int(d1*(float(C1.R))+d2*(float(C2.R)))
    if(R>255) :R=255
    G= int(d1*(float(C1.G))+d2*(float(C2.G)))
    if(G>255) :G=255
    B= int(d1*(float(C1.B))+d2*(float(C2.B)))
    if(B>255) :B=255
    return Color(R,G,B)
